local powerup_type = -1;
local powerup_haver = null;
local powerup_teammate = null;
local last_powerup_type = -1;
local prop_plr = Entities.FindByName(null, "powerup_prop_plr");
PrecacheScriptSound("Weapon_BarretsArm.Fizzle");

enum POWERUP_TYPES {
	POWERUP_DAMAGE,
	POWERUP_AGILITY,
	POWERUP_HASTE,
	POWERUP_RESISTANCE,
	POWERUP_PRECISION,
	POWERUP_REGEN
}

const POWERUP_TYPE_COUNT = 6;

local MODEL_ARRAY = ["models/pickups/pickup_powerup_strength.mdl",
					"models/pickups/pickup_powerup_agility.mdl",
					"models/pickups/pickup_powerup_haste.mdl",
					"models/pickups/pickup_powerup_resistance.mdl",
					"models/pickups/pickup_powerup_precision.mdl",
					"models/pickups/pickup_powerup_regen.mdl"];
					
local PlayerStatusTable = {};

// Folds all constants into global scope.
// https://developer.valvesoftware.com/wiki/Team_Fortress_2/Scripting/Script_Functions/Constants
foreach (a,b in Constants)
	foreach (k,v in b)
		if (!(k in getroottable()))
			getroottable()[k] <- v;

ClearGameEventCallbacks()

function OnGameEvent_player_death(params)
{
	local player = GetPlayerFromUserID(params.userid)

	PlayerStatusTable[player] <- false;
	EntFire("powerup_counter_players", "Subtract", 1, -1);
	
	if (player == powerup_haver){
		// The powerup prop should not be visible if the player's dead.
		EntFire("powerup_prop_plr", "Alpha", "0", -1);
	}
}

function OnGameEvent_player_spawn(params)
{
	local player = GetPlayerFromUserID(params.userid)
	
	if (PlayerStatusTable.rawin(player)){
		if (PlayerStatusTable[player] == false){
			// only update player counter if they were dead, otherwise do nothing
			EntFire("powerup_counter_players", "Add", 1, -1);
			EntFire(GetSpawnTriggerName(player), "Enable", "", -1);
		}
	}
	else{
		// if they don't exist, also update player counter if they exist and are not a spectator/unassigned.
		if (player.GetTeam() >= 2) {
			EntFire("powerup_counter_players", "Add", 1, -1);
			EntFire(GetSpawnTriggerName(player), "Enable", "", -1);
		}
	}
	if (player.GetTeam() >= 2) {
		// again, make sure they exist
		PlayerStatusTable[player] <- true;
	}
	
	if (player == powerup_haver){
		RemovePowerups();
	}
}


__CollectGameEventCallbacks(this)


function GeneratePowerup() {
	powerup_type = RandomInt(0, POWERUP_TYPE_COUNT - 1); // get our next, random, powerup
	EntFire("powerup_prop", "SetModel", MODEL_ARRAY[powerup_type], -1);
}

function ApplyPowerupToActivator(){
	// Apply the effects of the powerup to the activator - visual, gameplay, etc.
	last_powerup_type = powerup_type;
	powerup_haver = activator;
	
	// Weapon-specific modifiers - i.e. not "on wearer" modifiers.
	for (local ent; ent = Entities.FindByClassname(ent, "tf_weapon_*"); )
	{
		if (NetProps.GetPropEntity(ent, "m_hOwner") == powerup_haver){
			switch (last_powerup_type){
				case POWERUP_TYPES.POWERUP_DAMAGE:
					ent.AddAttribute("damage bonus", 1.2, -1); // +20% damage
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						ent.AddAttribute("Blast radius increased", 1.2, -1); // +20% blast radius
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						if (ent.GetClassname() == "tf_weapon_medigun"){
							ent.AddAttribute("heal rate bonus", 1.2, -1); // +20% heal rate
						}
					}
					break;
				case POWERUP_TYPES.POWERUP_HASTE:
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER || powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
						ent.AddAttribute("fire rate bonus", 0.8, -1); // +20% fire rate
						ent.AddAttribute("Reload time decreased", 0.8, -1); // +20% reload speed
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						ent.AddAttribute("Reload time decreased", 0.6, -1); // +40% reload speed					
					}
					break;
				case POWERUP_TYPES.POWERUP_PRECISION:
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						ent.AddAttribute("Projectile speed increased", 1.3, -1); // +30% projectile speed
						ent.AddAttribute("weapon spread bonus", 0.6, -1); // +40% more accurate
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT) {
						ent.AddAttribute("weapon spread bonus", 0.5, -1); // +50% more accurate
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						if (ent.GetClassname() == "tf_weapon_medigun"){
							ent.AddAttribute("ubercharge rate bonus", 1.25, -1); // +25% uber rate
							ent.AddAttribute("overheal decay bonus", 2.0, -1); // +60% longer overheal time
						}										
					}
					break;
				case POWERUP_TYPES.POWERUP_REGEN:
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						if (ent.GetClassname() == "tf_weapon_rocketlauncher"){
							ent.AddAttribute("health on radius damage", 15, -1); // +15hp on RL hit
						}
						else{
							ent.AddAttribute("heal on hit for rapidfire", 5, -1); // +5hp on any other hit
						}					
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
						if (ent.GetClassname() == "tf_weapon_scattergun"){
							ent.AddAttribute("heal on hit for rapidfire", 5, -1); // +5hp on scattergun hit
						}
						else{
							ent.AddAttribute("heal on hit for rapidfire", 15, -1); // +15hp on any other hit
						}					
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						ent.AddAttribute("heal on hit for rapidfire", 15, -1); // +15hp on hit
					}	
					break;
			}
		}
	}
	
	// "On wearer" modifiers - these can just be applied to the activator themselves.
	switch (last_powerup_type){
		case POWERUP_TYPES.POWERUP_AGILITY:
			powerup_haver.AddCustomAttribute("increased jump height", 1.75, -1); // +75% jump height
			powerup_haver.AddCustomAttribute("deploy time decreased", 0.4, -1); // +60% switch speed
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
				powerup_haver.AddCustomAttribute("move speed bonus", 1.4, -1); // +40% speed (336 ups)
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
				powerup_haver.AddCustomAttribute("move speed bonus", 1.3, -1); // +30% speed (520 ups)
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
				powerup_haver.AddCustomAttribute("move speed bonus", 1.125, -1); // +12.5% speed (360 ups)
				powerup_teammate = GetPowerupHolderTeammate();
				powerup_teammate.AddCustomAttribute("move speed bonus", 1.2, -1); // +20% speed (288 ups) for soldier
			}
			break;
		case POWERUP_TYPES.POWERUP_RESISTANCE:
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
				powerup_haver.AddCustomAttribute("rocket jump damage reduction", 0.7, -1); // -30% damage from rocketjumps
				powerup_haver.AddCustomAttribute("dmg taken increased", 0.90, -1); // -10% damage taken
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
				powerup_haver.AddCustomAttribute("dmg taken increased", 0.80, -1); // -20% damage taken
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
				powerup_haver.AddCustomAttribute("dmg taken increased", 0.85, -1); // -15% damage taken
				powerup_teammate = GetPowerupHolderTeammate();
				powerup_teammate.AddCustomAttribute("dmg taken increased", 0.85, -1); // for the soldier too
			}
			break;
		case POWERUP_TYPES.POWERUP_REGEN:
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
				powerup_haver.AddCustomAttribute("health from packs increased", 1.5, -1); // +50% hp from packs
				powerup_haver.AddCustomAttribute("max health additive bonus", 20, -1); // +20 max hp
				powerup_haver.AddCustomAttribute("ammo regen", 0.10, -1); // +10% ammo/5s
				powerup_haver.AddCustomAttribute("health regen", 5, -1); // +5 hp/s
			}
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
				powerup_haver.AddCustomAttribute("health from packs increased", 2, -1); // +100% hp from packs
				powerup_haver.AddCustomAttribute("health regen", 5, -1); // +5 hp/s
				powerup_haver.AddCustomAttribute("ammo regen", 0.10, -1); // +10% ammo/5s
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
				powerup_haver.AddCustomAttribute("health from packs increased", 1.5, -1); // +50% hp from packs
				powerup_haver.AddCustomAttribute("max health additive bonus", 10, -1); // +10 max hp
				powerup_haver.AddCustomAttribute("health regen", 3, -1); // +3 hp/s
			}
			break;
	}
	
	// Then, tint the activator's screen as per their team.
	ScreenFade(powerup_haver, 255, 255, 255, 255, 0.25, 0, 1); // flash screen team-coloured temporarily (this white blends with the team-coloured tint)
	if (powerup_haver.GetTeam() == TF_TEAM_RED){
		ScreenFade(powerup_haver, 255, 0, 0, 5, 0.02, 15, 0); // red
	}
	else {
		ScreenFade(powerup_haver, 0, 0, 255, 5, 0.02, 15, 0); // blue
	}
	EntFire("powerup_plr_fade", "Fade", "", -1);
	
	// Finally, put the powerup prop on top of the activator.
	prop_plr.SetModel(MODEL_ARRAY[powerup_type]);
	AddThinkToEnt(prop_plr, "PowerupPropThink");
}

function GetPowerupHolderTeammate(){
	// https://developer.valvesoftware.com/wiki/Team_Fortress_2/Scripting/VScript_Examples#Iterating_Through_Players
	for (local i = 1; i <= Constants.Server.MAX_PLAYERS; i++)
	{
		local player = PlayerInstanceFromIndex(i)
		if (player != null && player != powerup_haver) {
			if (player.GetTeam() == powerup_haver.GetTeam()){
				return player;
			}
		}
	}
}

function RemovePowerups(timer_expired = false){

	// Make sure the player's alive first!
	if (powerup_haver == null) {
		return;
	}
	
	// Weapon-specific modifiers - i.e. not "on wearer" modifiers.
	for (local ent; ent = Entities.FindByClassname(ent, "tf_weapon_*"); )
	{
		if (NetProps.GetPropEntity(ent, "m_hOwner") == powerup_haver){
			switch (last_powerup_type){
				case POWERUP_TYPES.POWERUP_DAMAGE:
					ent.RemoveAttribute("damage bonus");
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						ent.RemoveAttribute("Blast radius increased");
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						if (ent.GetClassname() == "tf_weapon_medigun"){
							ent.RemoveAttribute("heal rate bonus");
						}
					}
					break;
				case POWERUP_TYPES.POWERUP_HASTE:
					ent.RemoveAttribute("Reload time decreased");
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER || powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
						ent.RemoveAttribute("fire rate bonus");
						
					}
					break;
				case POWERUP_TYPES.POWERUP_PRECISION:
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						ent.RemoveAttribute("Projectile speed increased");
						ent.RemoveAttribute("weapon spread bonus");
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT) {
						ent.RemoveAttribute("weapon spread bonus");
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						if (ent.GetClassname() == "tf_weapon_medigun"){
							ent.RemoveAttribute("ubercharge rate bonus");
							ent.RemoveAttribute("overheal decay bonus");
						}										
					}
					break;
				case POWERUP_TYPES.POWERUP_REGEN:
					if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
						if (ent.GetClassname() == "tf_weapon_rocketlauncher"){
							ent.RemoveAttribute("health on radius damage");
						}
						else{
							ent.RemoveAttribute("heal on hit for rapidfire");
						}					
					}
					else if (powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT || powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
						ent.RemoveAttribute("heal on hit for rapidfire");			
					}
					break;
			}
		}
	}
	
	// "On wearer" modifiers - these can just be applied to the activator themselves.
	switch (last_powerup_type){
		case POWERUP_TYPES.POWERUP_AGILITY:
			powerup_haver.RemoveCustomAttribute("increased jump height");
			powerup_haver.RemoveCustomAttribute("deploy time decreased");
			powerup_haver.RemoveCustomAttribute("move speed bonus");
			if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
				powerup_teammate.RemoveCustomAttribute("move speed bonus");
			}
			break;
		case POWERUP_TYPES.POWERUP_RESISTANCE:
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER){
				powerup_haver.RemoveCustomAttribute("rocket jump damage reduction");
			}
			else if (powerup_haver.GetPlayerClass() == TF_CLASS_MEDIC){
				powerup_teammate.RemoveCustomAttribute("dmg taken increased");
			}
			powerup_haver.RemoveCustomAttribute("dmg taken increased");	
			break;
		case POWERUP_TYPES.POWERUP_REGEN:
			powerup_haver.RemoveCustomAttribute("health from packs increased");
			powerup_haver.RemoveCustomAttribute("health regen");
			powerup_haver.RemoveCustomAttribute("max health additive bonus");
			if (powerup_haver.GetPlayerClass() == TF_CLASS_SOLDIER || powerup_haver.GetPlayerClass() == TF_CLASS_SCOUT){
				powerup_haver.RemoveCustomAttribute("ammo regen");
			}
			break;
	}
	
	// Tell the other players that the powerup has expired.
	if (timer_expired && IsPlayerAlive(powerup_haver)) {
		PlayRemovePowerupSound();
	}
	
	// Be nice and clean up after ourselves.
	AddThinkToEnt(prop_plr, null);
	powerup_haver = null;
	powerup_teammate = null;
}

function PowerupPropThink(){
	// Continually update the prop's position and angles because parenting sucks
	local origin = powerup_haver.GetOrigin();
	origin.z += powerup_haver.GetPlayerMaxs().z + 4;
	self.SetOrigin(origin);
	self.SetAbsAngles(powerup_haver.GetAbsAngles());
	return -1;
}

function TeleportProjectile(is_tp1) {
	// Implement trigger_teleport but for projectiles, since TF2 doesn't support it for some reason.
	local origin = activator.GetOrigin();
	local landmark = null;
	local target = null;
	
	if (is_tp1) {
		landmark = Entities.FindByName(null, "tp1_proj"); // we can just use the origin of this entity unless something really fucks up hard
		target = Entities.FindByName(null, "tp1_target");
	}
	else {
		landmark = Entities.FindByName(null, "tp2_proj");
		target = Entities.FindByName(null, "tp2_target");
	}
	local offset = origin - landmark.GetOrigin();
	origin = target.GetOrigin() + offset;

	activator.Teleport(true, origin, false, QAngle(0,0,0), false, Vector(0,0,0));
}

function GetSpawnTriggerName(player) {
	// Create a format string instead of having 2x3 nested if statements.
	local team_str = null;
	local class_str = null;
	if (player.GetTeam() == TF_TEAM_RED){
		team_str = "red";
	}
	else {
		team_str = "blu";
	}
	
	switch (player.GetPlayerClass()) {
		case TF_CLASS_SOLDIER:
			class_str = "sl";
			break;
		case TF_CLASS_MEDIC:
			class_str = "md";
			break;
		case TF_CLASS_SCOUT:
			class_str = "sc";
			break;
	}
	
	return format("%s_spawn_%s", team_str, class_str);
}

function PlayRemovePowerupSound(){
	EmitSoundEx({
	sound_name = "Weapon_BarretsArm.Fizzle"
	origin = powerup_haver.GetOrigin()
	});
}

::IsPlayerAlive <- function(player)
{
	// https://developer.valvesoftware.com/wiki/Team_Fortress_2/Scripting/VScript_Examples#Checking_if_player_is_alive
    // lifeState corresponds to the following values:
    // 0 - alive
    // 1 - dying (probably unused)
    // 2 - dead
    // 3 - respawnable (spectating)
    return NetProps.GetPropInt(player, "m_lifeState") == 0;
}

// ---------------------------------------------------------------------
// LIST OF NUMBERS, FOR THOSE WHO COME ACROSS THIS FILE:
// NOTE: NO SCOUT NUMBERS; THIS MAP IS MAINLY BALANCED AROUND ULTIDUO.
// ---------------------------------------------------------------------
// With damage powerup (+20% damage/heal rate/blast radius):
//	- Rocket Launcher: 108 base, 29 min. splash, 135 max ramp-up, 175-unit splash radius
//	- Shotgun: 72 base, 108 max ramp-up
//	- Melee: 78 base
//  - Crossbow: 90-180 healing, 45-90 damage
//	- Medigun: 28.8hp/s heal rate, 86.4hp/s crit heals
//
// With agility powerup (S/M +40%/+12.5% movement speed, +75% jump height, +60% switch speed):
//	- Soldier: 336 ups; faster than default Medic speed. Same speed as Escape Plan at 41-79hp.
//	- Medic: 360 ups; as fast as a no-boost BFB scout. Brings Soldier to 288 ups, same speed as Escape Plan at 80-121hp; slightly faster than a Demo.
//	- Jump height: 120 units; can clear playermodels.
//	- Switch speed: All weapons can be switched to in 0.2 seconds.
//
// With haste powerup (S: +20% fire/reload speed, M: +40% reload speed):
//	- Can't do the maths right now, just imagine you're playing MVM with two pips into firing, one pip into reload for the Soldier,
//	  or for the Medic, two pips into reload.
//
// With resistance powerup:
//	- Soldier: TOTAL +40% rocket jump, 10% overall resistance; max 54 self-damage, 28 pencil-jump. Can *almost* facetank two rockets with no overheal.
//	- Medic: +15% damage resistance to both players; Soldier can facetank two rockets with no overheal, and Medic... survives better, I guess.
//
// With precision powerup (S: +30% projectile speed, -40% spread, M: +25% uber rate, +60% longer overheal decay):
//	- Soldier: Rockets are almost as fast as the Liberty Launcher, and his Shotgun has as good accuracy as the Shortstop.
//	- Medic: At full build, get 10% uber advantage over the enemy Medic across the 15 seconds of powerup duration. Overheal decays in 24 seconds
//    instead of 15.
//
// With regeneration powerup:
//	- Soldier: +5 hp/s regen; +10% ammo/5s regen - this is 2 rockets/3 shells every 5 seconds. Each RL hit heals up to 15hp; other weapons
//    replenish 5hp on hit. Health packs heal 60hp. Has 220hp.
// 	- Medic: +3 hp/s regen on top of natural regen, meaning 4-9 hp/s. Each hit heals up to 15hp. Health packs heal 45hp. Has 160hp.